Config Enforce Devel
====================

Config Enforce Devel is intended to be used on development sites to manage what configuration should be enforced by [Config Enforce](https://drupal.org/project/conig_enforce). This module should *only* be installed in development environments (See [Config Enforce Security Model](https://git.drupalcode.org/project/config_enforce/-/blob/1.0.x/README.md#security-model) for details).

Terminology
-----------

**Enforcement**: the goal of Config Enforce is to forcibly update active config from disk, to ensure that production config doesn't deviate from what's in code. This makes updating production config trivial, since we can update the code, refresh active config from there, and not worry about differences between them.

**Enforcement Level**: Config Enforce allows you to specify different levels of enforcement, depending on how rigidly you want to control the config. Currently there are 3 levels of enforcement:
  * **Allow form & API updates** will treat config objects as usual, installing them, but doesn't otherwise restrict changes in production or force updates from disk.
  * **Allow only API updates** will disable standard config forms from updating the config object, but allows programmatic updates to config, such as via `hook_update()` functions or the like.
  * **Read-only, no updates** further restricts *any* update to the config object entirely, including via API or programmatic updates. This enforcement level will also forcibly update active config with what's on disk when caches are rebuilt.

**Registry**: config enforce maintains a registry of enforced configs for each target module on a site, to track which config objects should be enforced, at what level, and in which config directory (`config/install` or `config/optional`). The registry also stores a hash of the config object itself, which is used to compare with active config and determine if a given config object needs updating.

**Target Module**: a module, theme, or install profile (any Drupal extension) that is configured within Config Enforce to contain enforced configs, and therefore has a registry and a set of `.yml` files it manages.


Getting Started
---------------

1. Enable `config_enforce_devel`, via the modules admin page, or via Drush: `drush pm-enable -y config_enforce_devel`
    1. This will enable `config_devel` (as a dependency), which will write any enforced configs to disk automatically, including Registries (see Terminology, above)
2. Visit the Enforced Configs page, available from the Toolbar menu or via the main menu Admin >> Configuration >> Development >> Config Enforce >> Enforced Configs (admin/config/development/config_enforce/enforced_configs).
    1. The first time that you visit this page, it will create a new Target Module for you called `config_enforce_default` (unless one already exists).
    2. This page contains:
        1. The names of all enforced configs, along with their enforcement settings. As a convenience, the names of config objects that were enforced from their config form will be rendered as links to their corresponding config forms.
        2. Buttons to "Generate from modules" or "Generate from active storage", which allow new config enforcement settings to be generated, either from previously exported config, or from the database (see Adding Config Enforce to an existing project, below)
        3. Controls for bulk operations that allow updates to, or deletion of, config enforcement settings for any selected configs.
3. Visit the "Settings" tab, or via the main menu *Admin >> Configuration >> Development >> Config Enforce >> Settings* (*admin/config/development/config_enforce/settings*)
    1. This page contains:
        1. An "Add module" button that will create a new target module for you.
        2. A list of modules available to set as target modules
        3. The default values to apply to new enforced configs
        4. A list of config objects to ignore from config enforcement.
    2. These settings are stored in a `config_enforce_devel.settings` config object, which you can enforce like any other config, so they will persist across reinstalls or developer environments. Typically you would store these in a "dev" target module which will not be enabled in production, since you won't have Config Enforce Devel there.


Basic Usage
-----------

You now have the basic setup of Config Enforce Devel, with one or more Target Modules configured, and some default enforcement settings in place. From here, you can proceed to go about your typical site building tasks, enforcing things as you go. As a simple example, to put the `system.site` config object under enforcement:

1. Go to *Admin >> Config >> System >> Basic site settings* (*admin/config/system/site-information*), and ensure the *Site Name*, *Email*, etc. are configured.
    1. You will see the Config Enforce "floaty box" widget appear in the top-right corner.
2. Check the Enforce config checkbox, and validate the *Target module*, *Directory*, and *Enforcement level* are correct (usually the defaults are fine, but you may want to move certain configs into different target modules).
3. Click Save configuration, and observe the form highlights the config object with a colour to indicate enforement level (green is read-only).
    1. If you look at the filesystem for your project, you will see a newly-written `system.site.yml` in the Target Module you selected, as well as an update to the `config_enforce.registry.*.yml` reflecting the newly-enforced config.
    2. Note that no `drush cex` or other manual intervention was required to export the config object- any subsequent changes to this config form will immediately be reflected on the filesystem.
4. `git add` and `git commit` both the `config_enforce.registry` as well as the `system.site` YAML files to capture them in your project repository.
5. Go back to the *Basic site settings* form and make a change, then use `git status` to observe the corresponding change immediately appear on the filesystem.
    1. Note that both the config object YAML file as well as the registry are updated with each change, since the content hash for the config object is updated to reflect the current state.

From here you can carry on with site-building tasks, checking "Enforce config" each time you get to a config form that hasn't yet been enforced, and periodically committing the various config objects and registries to reflect those changes.

Workflow example
----------------

To observe how the operational aspect of Config Enforce works, you can simulate a deployment with changes to the `system.site.yml`:

1. Ensure you have at least one config object enforced, eg. `system.site.yml`.
2. `drush pmu config_enforce_devel` to disable the developer UI, leaving only Config Enforce enabled.
3. Manually change the `system.site.yml` file, making a simple change to the *Site Name* or other string, and removing/altering the `hash` value so it differs from what's in active config
    1. Config Enforce Devel will automatically recalculate this hash each time it updates a config object, so when the new registry is deployed to production it will pick up the change
4. `drush cr` to rebuild caches, causing Config Enforce to re-scan all enforced configs and detect differing content hashes, then re-import any config objects which have changed.
5. Return to your site config form (eg. *Basic site settings*) and observe the simple string change has taken effect.
    1. Notice the config form itself is disabled (assuming you chose a high enough enforcement level), and you can't edit these settings directly anymore.
    2. The same principle applies to more complex config objects like Views, Page manager pages, Field settings, etc.


Adding Config Enforce to an existing project
--------------------------------------------

The first time you enable `config_enforce_devel` and go to *Enforced Configs*
(*admin/config/development/config_enforce/enforced_configs*), it will generate
a new `config_enforce_default` module to be the default target for enforced
configs. This is a convenience for bootstrapping a new Config Enforce setup. In
the case of an existing site, you probably already have some modules containing
configuration you'd like to enforce, and don't need the default target module.

To enable your existing modules, and remove the Config Enforce Default target,
follow these steps:

1. Go to *Enforced Configs > Settings* (*admin/config/development/config_enforce/settings*)
2. In the *Target modules* fieldset, select any of your existing modules and/or
   install profile containing exported configuration objects you want to enforce. Save configuration.
3. In the *Enforcement field defaults* fieldset, and select a different
   default Target module than *Config Enforce Default*. Save configuration.
4. Go back to the *Target modules* fieldset, and remove the *Config Enforce
   Default* module from the selected options. Save configuration.
5. Enforce these `config_enforce.devel` settings by checking the *Enforce
   Config* box in the Config Enforce "floaty window" UI. These should live in a
   `site_devel` target module which is only be turned on in development environments.
6. You can now remove the *Config Enforce Defaults* module from the
   *DRUPAL_ROOT/modules/custom* folder.

At this point Config Enforce Devel is configured with your existing set of
configuration-containing modules as targets for enforced configs. All that
remains is to generate enforcement settings for these exported configuration
objects.

To do so, there are two options: either a) generate enforcement settings from
the existing modules on disk, or b) generate enforcement settings by selecting
configuration in the active config (database) for the site.

### Generate from modules

From the main *Enforced Configs* admin screen, the *Generate from modules*
button will provide a list of all enabled Target modules, and nested lists of
all configuration objects exported to them.

Select an Enforcement level for the set of configuration objects you will
enforce. Select each of your Target modules in turn, and when selected, they
will expand to show the config directory (usually `optional` or `install`), and
finally the individual configuration objects within them.

Once you have selected everything you want to enforce, click Save. Config
Enforce Devel will write entries for each into a registry for the appropriate
target module, using the Enforcement level you selected. These config objects
are now enforced, and you can continue your development work.

### Generate from active storage

If there are configuration objects that are not yet in a target module on disk,
you can instead generate enforcement settings for them from Active config
storage.

In this case, when you click *Generate from active storage*, you will need to
select the Target module to save them into, as well as the Enforcement level
and config directory. The multi-select box below will show you all
configuration objects in Active storage that are not explicitly ignored by your
Config Enforce Devel settings (see below).

Once again, after selecting all the configuration objects you want to enforce,
and setting the Target module, Directory, and Enforcement level, and clicking
Save, these configuration objects will be written to disk in the Target module
you selected, and an entry is created for them in the Registry.

Suggested best practices
------------------------

1. Enforce and/or ignore every configuration object. This'll make it easier to identify when new config has been added.
2. TBD
